/* 
 *	Copyright (C) Aaron Holtzman - May 1999
 *
 *  This file is part of ac3dec, a free Dolby AC-3 stream decoder.
 *	
 *  ac3dec is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  ac3dec is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include "../global.h"
#include "ac3.h"
#include "bitstream.h"

void bit_allocate(uint_16 fscod, bsi_t *bsi, audblk_t *audblk);

// the floating point samples for one audblk
static stream_samples_t samples;

// the integer samples for the entire frame (with enough space for 2 ch out)
// if this size change, be sure to change the size when muting
static sint_16 s16_samples[2 * 6 * 256];

// storage for the syncframe
#define BUFFER_MAX_SIZE 4096
static uint_8 buffer[8][BUFFER_MAX_SIZE];
static sint_32 buffer_size[8];

uint_32 decode_buffer_syncframe(int channelIndex, syncinfo_t *syncinfo, uint_8 **start, uint_8 *end)
{
	uint_8 *cur = *start;
	uint_16 syncword = syncinfo->syncword;
	uint_32 ret = 0;

	// find an ac3 sync frame
resync:
	while(syncword != 0x0b77)
	{
		if(cur >= end)
			goto done;
		syncword = (syncword << 8) + *cur++;
	}

	// need the next 3 bytes to decide how big the frame is
	while(buffer_size[channelIndex] < 3)
	{
		if(cur >= end)
			goto done;

		buffer[channelIndex][buffer_size[channelIndex]++] = *cur++;
	}
	
	parse_syncinfo(syncinfo, &buffer[channelIndex][0]);

	if (syncinfo->frame_size==0)		// CRITICAL CONDITION
		goto done;

	while (buffer_size[channelIndex] < (syncinfo->frame_size<<1) - 2)
	{
		if(cur >= end)
			goto done;

		buffer[channelIndex][buffer_size[channelIndex]++] = *cur++;
	}

	// check the crc over the entire frame
	if (crc_process_frame(&buffer[channelIndex][0], (syncinfo->frame_size<<1) - 2))
	{
		syncword = 0xffff;
		buffer_size[channelIndex] = 0;
		goto resync;
	}

	// if we got to this point, we found a valid ac3 frame to decode
	bitstream_init(&buffer[channelIndex][0]);
	// get rid of the syncinfo struct as we already parsed it
	bitstream_get(24);

	// reset the syncword for next time
	syncword = 0xffff;
	buffer_size[channelIndex]= 0;
	ret = 1;

done:
	syncinfo->syncword = syncword;
	*start = cur;
	return ret;
}

uint_32 ac3_decode_data(int channelIndex, uint_8 * data_start, int size)
{
	uint_8 *data_end = data_start + size;
	uint_32 i, j=0;

	curChannelIndex = channelIndex;

	while (decode_buffer_syncframe(channelIndex, &channelData[channelIndex].syncinfo, &data_start, data_end))
	{
		if (channelData[channelIndex].error_flag)
		{
			ZeroMemory(s16_samples, sizeof(sint_16) * 256 * 2 * 6);
			channelData[channelIndex].error_flag = 0;
			continue;
		}

		parse_bsi(&channelData[channelIndex].bsi);

		for (i=0; i<6; i++)
		{
			ZeroMemory(samples, sizeof(double) * 256 * (channelData[channelIndex].bsi.nfchans + channelData[channelIndex].bsi.lfeon));

			parse_audblk(&channelData[channelIndex].bsi, &channelData[channelIndex].audblk);

			exponent_unpack(&channelData[channelIndex].bsi, &channelData[channelIndex].audblk); 
			if (channelData[channelIndex].error_flag)
				goto error;

			bit_allocate(channelData[channelIndex].syncinfo.fscod, &channelData[channelIndex].bsi, &channelData[channelIndex].audblk);

			coeff_unpack(&channelData[channelIndex].bsi, &channelData[channelIndex].audblk, samples);
			if (channelData[channelIndex].error_flag)
				goto error;

			if (channelData[channelIndex].bsi.acmod == 0x2)
				rematrix(&channelData[channelIndex].audblk, samples);

			imdct(&channelData[channelIndex].bsi, &channelData[channelIndex].audblk, samples);

			downmix(&channelData[channelIndex].audblk, &channelData[channelIndex].bsi, samples, (sint_16 *)(audioData[channelIndex].data+i*512*2));

			sanity_check(&channelData[channelIndex].bsi, &channelData[channelIndex].audblk);
			if (channelData[channelIndex].error_flag)
				goto error;
		}
		audioData[channelIndex].data+=6144;

		//memcpy(&AC3Dec_Buffer[start], s16_samples, 6144);
		//start += 6144;
		j++;
error:
		;
	}

	audioData[channelIndex].bytesUsed+=j*6144;
	return j*6144;
}

void InitialAC3(int channelIndex)
{
	curChannelIndex = channelIndex;

	channelData[channelIndex].error_flag = 0;
	buffer_size[channelIndex] = 0;

	ZeroMemory(&channelData[channelIndex].syncinfo, sizeof(syncinfo_t));
	ZeroMemory(&channelData[channelIndex].bsi, sizeof(bsi_t));
	ZeroMemory(&channelData[channelIndex].audblk, sizeof(audblk_t));

	drc_init();
	imdct_init();
	exponent_init();
	mantissa_init();

	srand(0);
}
