/* 
 *  Deinterlacer.cpp - Original code by Donald Graft
 *                                      neuron2@home.com.
 *
 *	Copyright (C) Alberto Vigata - January 2000
 *
 *  This file is part of FlasKMPEG, a free MPEG to MPEG/AVI converter
 *	
 *  FlasKMPEG is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  FlasKMPEG is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <windows.h>
#include <vfw.h>
#include "flask.h"
#include "vbitmap.h"
#include "Deinterlacer.h"

struct MyFilterData {
	int					*prevFrame;
	unsigned char		*prevFrameY;
	unsigned char		*prevFrameV;
	unsigned char		*prevFrameU;
	unsigned long		*moving;
	int 				Blend;
	int 				threshold;
	int					bitpitch;
	int					bitpitchY;
	bool				skip;
	int					first;
} my_mfd;

MyFilterData *mfd = &my_mfd;




int StartDeinterlacer(int w, int h, int blend, int threshold) {

	mfd->prevFrame		= new int[w*h];
	mfd->Blend          = blend;
	mfd->threshold      = threshold;

	memset(mfd->prevFrame, 0, w*h*sizeof(int));


	mfd->bitpitch		= (w + 32 - 1) / 32;
	mfd->moving			= new unsigned long[mfd->bitpitch * h];
	memset(mfd->moving, 0, sizeof(unsigned long)*mfd->bitpitch * h);
	mfd->skip = FALSE;
	mfd->first = TRUE;
	
	return 0;
}

#define ASM_COPY(o,i,runsize)  __asm             \
{									\
__asm			mov ecx, runsize    \
__asm			mov esi, i          \
__asm           mov edi, o          \
__asm			rep movsd           \
}



int DeInterlace(FRAMEDATA *source, FRAMEDATA *dest) {

	const int		bitpitch = mfd->bitpitch;
	const long		pitch = source->stride;
	const PixDim	w = source->width;
	const PixDim	h = source->height;
	Pixel32 *		src = (Pixel32 *) source->data;
	Pixel32 *		dst = (Pixel32 *) dest->data;

	long sourcePitch = source->stride;
	long destPitch = dest->stride;

	if(!source || !dest)
		return 0;

	int x, y;
	long prevValue;
	Pixel32 p0, p1, p2;
	long r, g, b, r0, g0, b0;
	int *lumptr;

	if (h<2) return 0;

	// Compute differences for all pixels by comparing each pixel
    // to its corresponding pixel in the previous frame. Then do a threshold
    // test of the difference to decide if the pixel is 'moving'.
	// Then create an array of flags indicating pixels that have moved since
	// the last frame.

	lumptr = mfd->prevFrame;
	unsigned long *maskptr = mfd->moving + mfd->bitpitch;

	for (y = 1; y < h - 1; y++)
	{
		long pixmask = 0;

		src = (Pixel *)((char *)src + sourcePitch);

		x = 0;
		do
		{
			// Set the moving flag if the diff exceeds the configured
            // threshold.
			pixmask <<= 1;
			prevValue = *lumptr;
			*lumptr++ = src[x];
			b = (src[x] & 0xff);
			b0 = (prevValue & 0xff);
			if (abs(b - b0) > mfd->threshold)
				goto moving;
			r = (src[x] & 0xff0000) >> 16;
			r0 = (prevValue & 0xff0000) >> 16;
			if (abs(r - r0) > mfd->threshold)
				goto moving;
			g = (src[x] & 0xff00)   >> 8;
			g0 = (prevValue & 0xff00)   >> 8;
			if (abs(g - g0) > mfd->threshold)
				goto moving;			
			goto notmoving;
moving:
			pixmask |= 1;
notmoving:
			if (!(++x & 31))
				*maskptr++ = pixmask;
		} while(x<w);

		if (x & 31)
			*maskptr++ = pixmask << (-x&31);
	}

	// Remove combing from the motion areas and render.
    // The first line gets a free ride.
	src = (Pixel32 *) source->data;
	dst = (Pixel32 *) dest->data;
	//memcpy(dst, src, w*4);

	ASM_COPY(dst, src, w);

	maskptr = mfd->moving;

	for (y = 1; y < h - 1; y++)
	{
		unsigned long mask;

		src = (Pixel *)((char *)src + sourcePitch);
		dst = (Pixel *)((char *)dst + destPitch);

		if (mfd->Blend) {
			x = 0;
			do {
				if (!(x & 31)) {
					mask = maskptr[bitpitch] | maskptr[0] | maskptr[bitpitch*2];
					++maskptr;
				}

				if ((signed long)mask >= 0)
					dst[x] = src[x];
				else {

					/* Blend fields. */
					p0 = src[x];
					p0 &= 0x00fefefe;

					p1 = ((Pixel32 *)(((char *)src)-pitch))[x];
					p1 &= 0x00fcfcfc;

					p2 = ((Pixel32 *)(((char *)src)+pitch))[x];
					p2 &= 0x00fcfcfc;

					dst[x] = (p0>>1) + (p1>>2) + (p2>>2);
				}
				mask <<= 1;

			} while(++x<w);
		} else {
			// Doing line interpolate. Thus, even lines are going through
			// for moving and non-moving mode. Odd line pixels will be subject
			// to the motion test.
			if (y&1) {
				x = 0;
				do {
					if (!(x & 31)) {
						mask = maskptr[bitpitch] | maskptr[0] | maskptr[bitpitch*2];
						++maskptr;
					}

					if ((signed long)mask >= 0)
						dst[x] = src[x];
					else {
						p1 = ((Pixel32 *)(((char *)src)-pitch))[x];
						p1 &= 0x00fefefe;

						p2 = ((Pixel32 *)(((char *)src)+pitch))[x];
						p2 &= 0x00fefefe;

						dst[x] = (p1>>1) + (p2>>1);
					}

					mask <<= 1;

				} while(++x<w);
			} else {
				// Even line; pass it through.
				//memcpy(dst, src, w*4);

				ASM_COPY(dst, src, w);

				maskptr += bitpitch;
			}
		}


	}
	
	// The last line gets a free ride.
	src = (Pixel *)((char *)src + sourcePitch);
	dst = (Pixel *)((char *)dst + destPitch); //MAF
	ASM_COPY(dst, src, w);

	return 0;
}

int StopDeinterlacer() {

	delete[] mfd->prevFrame;	
	mfd->prevFrame = NULL;
	delete[] mfd->moving;		
	mfd->moving = NULL;

	return 0;
}

int StartDeinterlacerYv12(int w, int h, int blend, int threshold) {

	mfd->prevFrameY		= new unsigned char [w*h*2];
	mfd->prevFrameV = mfd->prevFrameY + (w * h);
	mfd->prevFrameU = mfd->prevFrameV + ((w/2) * (h/2));
	memset(mfd->prevFrameY, 0, w*h*2);


	mfd->Blend          = blend;
	mfd->threshold      = threshold;

	mfd->bitpitchY		= (w + 32 - 1) / 32;

	mfd->moving			= new unsigned long[mfd->bitpitchY * h];

	memset(mfd->moving, 0, sizeof(unsigned long)*mfd->bitpitch * h);
	mfd->skip = FALSE;
	mfd->first = TRUE;
	
	return 0;
}

int DeInterlaceYv12(FRAMEDATA *source, FRAMEDATA *dest) {

	const int		bitpitchY = mfd->bitpitchY;

	const long		pitchY = source->stride;
	const long		pitchV = source->stride/2;
	const long		pitchU = source->stride/2;

	const PixDim	w = source->width;
	const PixDim	w2 = source->width/2;
	const PixDim	w4 = source->width/4;
	const PixDim	w8 = source->width/8;
	const PixDim	h = source->height;
	unsigned char *		srcY = (unsigned char *) source->data;
	unsigned char *		srcV = srcY + (w * h);
	unsigned char *		srcU = srcV + ((w/2) * (h/2));

	unsigned char *		dstY = (unsigned char *) dest->data;
	unsigned char *		dstV = dstY + (w * h);
	unsigned char *		dstU = dstV + ((w/2) * (h/2));

	long sourcePitchY = source->stride;
	long sourcePitchV = source->stride/2;
	long sourcePitchU = source->stride/2;

	long destPitchY = dest->stride;
	long destPitchV = dest->stride/2;
	long destPitchU = dest->stride/2;

	if(!source || !dest)
		return 0;

	int xx, yy;
	unsigned char prevValueY;
	unsigned char prevValueU;
	unsigned char prevValueV;
	unsigned char p0Y, p1Y, p2Y;
	unsigned char p0V, p1V, p2V;
	unsigned char p0U, p1U, p2U;
	unsigned char y, y0, v,v0,u,u0;
	unsigned char *lumptrY;
	unsigned char *lumptrV;
	unsigned char *lumptrU;



	if (h<2) return 0;

	// Compute differences for all pixels by comparing each pixel
    // to its corresponding pixel in the previous frame. Then do a threshold
    // test of the difference to decide if the pixel is 'moving'.
	// Then create an array of flags indicating pixels that have moved since
	// the last frame.

	lumptrY = mfd->prevFrameY;
	lumptrV = mfd->prevFrameV;
	lumptrU = mfd->prevFrameU;

	unsigned long *maskptr = mfd->moving + mfd->bitpitchY;

	for (yy = 1; yy < h - 1; yy++)
	{
		long pixmask = 0;

		srcY = (unsigned char *)(srcY + sourcePitchY);

		if(!(yy&0x01))
			{
			srcV = (unsigned char *)(srcV + sourcePitchV);
			srcU = (unsigned char *)(srcU + sourcePitchU);

			dstV = (unsigned char *)(dstV + destPitchV);
			dstU = (unsigned char *)(dstU + destPitchU);
			}

		xx = 0;
		do
		{
			// Set the moving flag if the diff exceeds the configured
            // threshold.
			pixmask <<= 1;
			prevValueY = *lumptrY;
			*lumptrY++ = srcY[xx];
			y = srcY[xx];
			y0 = prevValueY;
/*
			if(yy&0x01 && xx&0x01)
				{
				prevValueV = *lumptrV;
				*lumptrV++ = srcV[xx/2];
				v = srcV[xx/2];
				v0 = prevValueV;

				prevValueU = *lumptrU;
				*lumptrU++ = srcU[xx/2];
				u = srcU[xx/2];
				u0 = prevValueU;
				
				if (abs(v - v0) > (mfd->threshold))
					goto moving;

				if (abs(u - u0) > (mfd->threshold))
					goto moving;
				}
*/

			if (abs(y - y0) > (mfd->threshold))
				goto moving;

			goto notmoving;
moving:
			pixmask |= 1;
notmoving:
			if (!(++xx & 31))
				*maskptr++ = pixmask;
		} while(xx<w);

		if (xx & 31)
			*maskptr++ = pixmask << (-xx&31);
	}

	// Remove combing from the motion areas and render.
    // The first line gets a free ride.
	srcY = (unsigned char *) source->data;
	srcV = srcY + (w * h);
	srcU = srcV + ((w/2) * (h/2));

	dstY = (unsigned char *) dest->data;
	dstV = dstY + (w * h);
	dstU = dstV + ((w/2) * (h/2));
	//memcpy(dst, src, w*4);

	ASM_COPY(dstY, srcY, w4);
	ASM_COPY(dstV, srcV, w8);
	ASM_COPY(dstU, srcU, w8);

	maskptr = mfd->moving;

	for (yy = 1; yy < h - 1; yy++)
	{
		unsigned long mask;

		srcY = (unsigned char *)(srcY + sourcePitchY);
		dstY = (unsigned char *)(dstY + destPitchY);

		if(!(yy&0x01))
			{
			srcV = (unsigned char *)(srcV + sourcePitchV);
			srcU = (unsigned char *)(srcU + sourcePitchU);

			dstV = (unsigned char *)(dstV + destPitchV);
			dstU = (unsigned char *)(dstU + destPitchU);
			}

		if (mfd->Blend) {
			xx = 0;
			do {
				if (!(xx & 31)) {
					mask = maskptr[bitpitchY] | maskptr[0] | maskptr[bitpitchY*2];
					++maskptr;
				}

				if ((signed long)mask >= 0)
					{
					dstY[xx] = srcY[xx];
					if(yy&0x01 && xx&0x01)
						{
						dstV[xx/2]=srcV[xx/2];
						dstU[xx/2]=srcU[xx/2];
						}

					}
				else {

					/* Blend fields. */
					p0Y = srcY[xx];
					p1Y = ((unsigned char *)(((unsigned char *)srcY)-pitchY))[xx];
					p2Y = ((unsigned char *)(((unsigned char *)srcY)+pitchY))[xx];
					dstY[xx] = (p0Y>>1) + (p1Y>>2) + (p2Y>>2);

					if(yy&0x01 && xx&0x01)
						{
						p0V = srcV[xx/2];
						p2V = ((unsigned char *)(((unsigned char *)srcV)+pitchV))[xx/2];
						dstV[xx/2] = (p0V>>1) + (p0V>>2) + (p2V>>2);

						p0U = srcU[xx/2];
						p2U = ((unsigned char *)(((unsigned char *)srcU)+pitchU))[xx/2];
						dstU[xx/2] = (p0U>>1) + (p0U>>2) + (p2U>>2);
						}
				}
				mask <<= 1;

			} while(++xx<w);
		} else {
			// Doing line interpolate. Thus, even lines are going through
			// for moving and non-moving mode. Odd line pixels will be subject
			// to the motion test.


			
			if (yy&1) {
				xx = 0;

				do {
					if (!(xx & 31)) {
						mask = maskptr[bitpitchY] | maskptr[0] | maskptr[bitpitchY*2];
						++maskptr;
					}

					if ((signed long)mask >= 0)
						{
						dstY[xx] = srcY[xx];
						if(xx&0x01)
							{
							dstV[xx/2]=srcV[xx/2];
							dstU[xx/2]=srcU[xx/2];
							}
						}
					else {
						p1Y = ((unsigned char *)(((unsigned char *)srcY)-pitchY))[xx];
						p2Y = ((unsigned char *)(((unsigned char *)srcY)+pitchY))[xx];

						dstY[xx] = (p1Y>>1) + (p2Y>>1);
						if(xx&0x01)
							{
							p1V = ((unsigned char *)(((unsigned char *)srcV)))[xx/2];
							p2V = ((unsigned char *)(((unsigned char *)srcV)+pitchV))[xx/2];
							dstV[xx/2] = (p1V>>1) + (p1V>>2) + (p2V>>2);

							p1U = ((unsigned char *)(((unsigned char *)srcU)))[xx/2];
							p2U = ((unsigned char *)(((unsigned char *)srcU)+pitchU))[xx/2];
							dstU[xx/2] = (p1U>>1) + (p1U>>2) + (p2U>>2);

							}
					}

					mask <<= 1;

				} while(++xx<w);
			} else {
				// Even line; pass it through.
				//memcpy(dst, src, w*4);

				ASM_COPY(dstY, srcY, w4);
				ASM_COPY(dstV, srcV, w8);
				ASM_COPY(dstU, srcU, w8);

				maskptr += bitpitchY;
			}
			
		}


	}
	
	// The last line gets a free ride.
	srcY = (unsigned char *)(srcY + sourcePitchY);
	srcV = (unsigned char *)(srcV + sourcePitchV);
	srcU = (unsigned char *)(srcU + sourcePitchU);

	dstY = (unsigned char *)(dstY + destPitchY);
	dstV = (unsigned char *)(dstV + destPitchV);
	dstU = (unsigned char *)(dstU + destPitchU);
	ASM_COPY(dstY, srcY, w4);
	ASM_COPY(dstV, srcV, w8);
	ASM_COPY(dstU, srcU, w8);

	return 0;
}
